/* 
Reziproke Frequenzmessung mit STM32F730 mit 8-stelliger Auflsung.
Eingangsfrequenbereich: 0,1 Hz - 100 MHz

Die Frequenz wird aus den Eingangsimpulsen/Zeit errechnet. Dazu werden die 
Timer9 und Timer10 verwendet. Als Zeitbasis fr alle Messungen dient die
interne Frequenz SystemCoreClock mit (hier) 216 MHz. 

Timer9 zhlt am Eingang TIM9_CH1 (PE5) forlaufend die Eingangsimpulse. 
Timer10 zhlt fortlaufend die interen Impulse von SystemCoreClock (216 MHz).
Beide Zhler werden einmalig gestartet und dann weder gestoppt noch gelscht.
Timer6 erzeugt das Timing fr die Messdauer im 1 ms Raster.

Ein ext. D-FF erzeugt auf Anforderung mit Signal TRIG2 (PC0) synchron zu einem 
Eingangsimpuls das Signal CAP2 (TIM9_CH2 (PE6) fr Timer9 und TIM10_CH1 (PB8) 
fr Timer10). In den zugehrigen Capture-Registern werden die Anzahl der Ereignisse
sowie deren genauer Zeitpunkt festgehalten. Zustzlich werden auch die berlufe
der Timer bercksichtigt, soda beide Messwerte als uint32_t Variablen zur
Auswertung bereitstehen: "F2_end_ereignis" und "F2_end_zeit".
Diese Zeitpunkte werden in der ISR TIM1_BRK_TIM9_IRQHandler ermittelt.

Ein einzelner Messpunkt liefert noch kein Ergebnis, sondern es mssen ein 
Startzeitpunkt und ein Endzeitpunkt vorliegen, um aus deren Differenz Ereignisse
und Zeit zu erhalten: "F2_mess_ereignisse" und "F2_mess_zeit".

"F2_mess_zeit" enthlt die Anzahl der internen Referenzimpulse und mu daher noch
mit Division durch SystemCoreClock auf Sekunde umgerechnet werden. Das Ergebnis
ist dann "eff_zeit2".
Das Messergebnis ergibt sich dann als: "F2_ergebnis = F2_mess_ereignisse / eff_zeit2".

Nach einer Auswertung werden die letzten Endpunkte zu den Startpunkten fr das
neue Intervall.

Wie geschrieben wird eine Auswertung immer mit dem Signal TRIG2 angefordert und
synchron zu einem Eingangsimpuls gestartet. Wenn man schnell viele Messwerte
braucht, knnte man nach jeder Auswertung mit TRIG2 gleich eine neue Auswertung
anfordern. Der hier verwendete C STM32F730 schafft so einige 10000 Messungen/s,
sofern man auf die ser. Datenausgabe verzichtet.
Bentigt man jedoch eine mglichst hohe Auflsung der Messwerte, mu man dafr
sorgen, da die Zeitmessung mglichst hoch aufgelst ist. Dies gilt nicht fr 
die Eingangsimpulse, da diese den Zeitpunkt der Auswertung synchron vorgeben.

Die Zeitmessung luft mit der internen Frequenz SystemCoreClock. Die hier 
eingestellte Maximalfrequenz betrgt 216 MHz oder anders geschrieben 2.16E08 Hz.
Am Exponenten ist zu erkennen, da in einer Sekunde eine 8-stellige Auflsung
bei der Zeitmessung vorliegt. Selbst bei 1/2 s Messzeit sind mit 1.08E08
Referenzimpulsen noch 8-stellige Werte erreichbar.
Fr ein 8-stelliges Messergebnis sind die Auswertungen mit einem Abstand von
>= 0,5 s anzufordern.
Verkrzt man diese Zeit auf z.B. 1 ms, geht die Auflsung auf 5 Stellen/Messwert
zurck. Da jede Auswertung synchron zu einem Eingangsimpuls gestartet wird,
wird die Messrate bei niedrigen Frequenzen reduziert. Bei 1 Hz Eingangssignal
ist nur 1 Messung/s mglich.
Fehlt das Eingangssignal ganz, wird per Timeout die Messung abgebrochen und
neu gestartet.

Die Messwerte werden ber USART3 ausgegeben: "sende_sting(s)"

2020-10-12
Michael Nowak
http://www.mino-elektronik.de
Alle Angaben ohne Gewaehr !
*/

#define F2_MESSUNG

#include <stdio.h>
#include <string.h>
#include "stm32f7xx.h"
extern void sende_string(char *s);    // Ausgabe ber USART3

#define DEF_F2_TAKT       216000000   // 216 MHz
#define DEF_F2_MESSZEIT   666         // 0,666 s
#define DEF_TIMEOUT       10000       // nach 10 s
#define MAX_ZEICHEN       50          // fuer Ergebnis String

#define PRIO_TIMER        4           // hohe Prioritt fr T9 und T10
#define PRIO_TIM6         14          // niedrig, unkritisch
#define AFR_TIM9          3           // fr T9  lt. Datenblatt
#define AFR_TIM10         3           // fr T10 lt. Datenblatt


#define BIT(x) (1<<x)
enum gpio_modus {IN, OUT, AF, ANALOG};
enum gpio_speed {LOW_SPEED, MEDIUM_SPEED, HIGH_SPEED, V_HIGH_SPEED};
enum gpio_pullup {NO_PULL, PULL_UP, PULL_DOWN};

#define CLR_MODER         3         // Lschmaske
#define SET_GPIO          0x1       // untere Bits
#define RESET_GPIO        0x10000   // obere Bits

// PortB
#define TIM10_CAPTURE2    8
// PortC
#define TRIG2             0
// PortE
#define TIM9_ZAEHLER2     5
#define TIM9_CAPTURE2     6

enum ablauf {MESSEN=0, AUSLESEN, AUSWERTEN};        // die Phasen der Messung
 
double              F2_ergebnis,                    // in Hz
                    eff_zeit2;                      // eff. Messzeit in s
double              F2_ref_frequenz = DEF_F2_TAKT;  // Vorgabewerte
int32_t             F2_ref_offset = -250,              // fr Abgleich von F2_ref
                    F2_messzeit = DEF_F2_MESSZEIT,
                    F2_timeout = DEF_TIMEOUT;

volatile uint8_t    F2_messwert_vorhanden,          // gueltige Messungen
                    F2_mess_status;                 // Ablaufsteuerung
                                   
volatile uint32_t   F2_mess_dauer;                  // in ms
uint32_t            F2_start_ereignis,              // Impulse zu Beginn der Messung
                    F2_start_zeit,                  // rel. Zeit: Beginn der Messung
                    F2_zeit_ueberlauf,
                    F2_end_zeit,                    // Zeitpunkt der Auswertung
                    F2_mess_zeit,                   // genaue Zeit der Messung
                    F2_mess_ereignisse,             // Impulse der Messung
                    F2_ueberlauf; 
volatile uint32_t   F2_end_ereignis;                // Impulse am Ende der Messung


// direktes Setzen oder Ruecksetzen eines Port-Pins
void set_pin(GPIO_TypeDef* GPIOx, uint8_t number, char set) {
  if(set) {
    GPIOx->BSRR = (SET_GPIO << number);       // IO-Bit setzen
  } else {
    GPIOx->BSRR = (RESET_GPIO << number);     // IO-Bit loeschen
  }
}


// 1 kHz ISR mit T6 erzeugen
void init_T6(void)
{
static uint8_t init;
  if(!init) {
    RCC->DCKCFGR1 |= RCC_DCKCFGR1_TIMPRE;       // Timer auf max. Takt
    init = 1;
    RCC->APB1ENR |= RCC_APB1ENR_TIM6EN;
    TIM6->ARR = 1000-1;                         // Teiler fr 1 kHz
    TIM6->PSC = SystemCoreClock/1000000-1;      // Vorteiler auf 1 MHz
    TIM6->CR1 = 0x1;                            // starten
    TIM6->DIER = 0x1;                           // und Interrupt freigeben
    
    NVIC_EnableIRQ(TIM6_DAC_IRQn);
    NVIC_SetPriority(TIM6_DAC_IRQn,PRIO_TIM6);  // geringe Prioritaet
  }    
} 

// Messdauer im 1 ms Raster inkrementieren
void TIM6_DAC_IRQHandler(void)              // Aufruf mit 1 kHz
{
  TIM6->SR = 0;                             // flag loeschen
  F2_mess_dauer++;                          // Mesdauer in ms-Raster
  __ISB();                                  // gelschte Flags abwarten
}    


void init_t9(void)
{
  RCC->APB2ENR |= RCC_APB2ENR_TIM9EN;
	RCC->AHB1ENR |= (RCC_AHB1ENR_GPIOEEN);    // PortE aktivieren
  TIM9->CR1 = 0;                            // T9 stoppen, falls schon aktiv
  
// alternative Funktion der Zaehlereingaenge einschalten  
  GPIOE->MODER |= (AF << (TIM9_ZAEHLER2 * 2) |
                   AF << (TIM9_CAPTURE2 * 2));                       
  
// Pullup zu den Eingaengen einschalten
  GPIOE->PUPDR |= (PULL_UP << (TIM9_ZAEHLER2 * 2) |
                   PULL_UP << (TIM9_CAPTURE2 * 2));                       

// PE5 als Zhlereingang, PE6 als capture-input zu T9
  GPIOE->AFR[0] |=  (AFR_TIM9 << (TIM9_ZAEHLER2 * 4) | 
                     AFR_TIM9 << (TIM9_CAPTURE2 * 4));                   


  NVIC_EnableIRQ(TIM1_BRK_TIM9_IRQn);                   // Int-T9 aktivieren
  NVIC_SetPriority(TIM1_BRK_TIM9_IRQn,PRIO_TIMER);

  TIM9->CCMR1 = TIM_CCMR1_CC2S_0;                       // capture zuordnung  IC2 -> TI2
  TIM9->CCER =  TIM_CCER_CC2E;                          // pos. Flanken an IC2 freigeben
  TIM9->SMCR |= TIM_SMCR_TS_2 + TIM_SMCR_TS_0;          // ext. Signal von TI1
  TIM9->SMCR |= TIM_SMCR_SMS_2 + TIM_SMCR_SMS_1 + TIM_SMCR_SMS_0;   // Zhlereingang von TI1

  TIM9->DIER |= TIM_DIER_CC2IE + TIM_DIER_UIE;          // mit interrupts
  TIM9->CR1 |=  TIM_CR1_CEN;                            // T9 starten  
}

void init_t10(void)
{
  RCC->APB2ENR |= RCC_APB2ENR_TIM10EN;
	RCC->AHB1ENR |= (RCC_AHB1ENR_GPIOBEN);    // PortB aktivieren
  TIM10->CR1 = 0;                           // T10 stoppen, falls schon aktiv

  GPIOB->MODER |= (AF << (TIM10_CAPTURE2 * 2));         // alternative Funktion         
  GPIOB->PUPDR |= (PULL_UP << (TIM10_CAPTURE2 * 2));    // Pullup einschalten

// PB8 als capture-input zu T10
  GPIOB->AFR[1] |=  (AFR_TIM10 << ((TIM10_CAPTURE2-8) * 4));                   


  NVIC_EnableIRQ(TIM1_UP_TIM10_IRQn);                   // Int-T10 aktivieren
  NVIC_SetPriority(TIM1_UP_TIM10_IRQn,PRIO_TIMER);
  
  TIM10->CCMR1 =  TIM_CCMR1_CC1S_0;                     // capture zuordnung  IC1 -> TI1
  TIM10->CCER = TIM_CCER_CC1E;                          // pos. Flanken an IC1 freigeben

  TIM10->DIER |= TIM_DIER_UIE;                          // mit interrupts
  TIM10->CR1 |= TIM_CR1_CEN;                            // T10 starten
}

// ISR zu Timer10
// berlufe bei der Zeitmessung zhlen
void TIM1_UP_TIM10_IRQHandler(void)
{
  if(TIM10->SR & TIM_SR_UIF) {                          // bei berlufen von T10
     TIM10->SR = ~TIM_SR_UIF;                           // flag wieder loeschen
     F2_zeit_ueberlauf += 0x10000;                      // und berlufe zhlen
  }
  __ISB();                                    // gelschte Flags abwarten
}

// ISR zu Timer9
// synchron zu einem Eingangsimpuls Ereignisse und Zeitpunkt ermitteln
// anschlieend Auswertung freigeben
// ferner berlaufe bei der Ereigniszhlung zhlen
void TIM1_BRK_TIM9_IRQHandler(void)
{
uint32_t temp;
  if(TIM9->SR & TIM_SR_CC2IF) {                         // bei Capture-Ereignis
    TIM9->SR = ~TIM_SR_CC2IF;                           // flag wieder loeschen
    temp = TIM9->CCR2;                                  // Capture-Wert Ereignisse
    F2_end_ereignis = F2_ueberlauf + temp;              // Endwert Ereignisse
    if((TIM9->SR & TIM_SR_UIF) && (temp < 0x8000))      // evtl. Ueberlauf T9 noch offen?
      F2_end_ereignis += 0x10000;                       // nur, wenn capture-int + overflow-int gleichzeitig !

    temp = TIM10->CCR1;                                 // Capture-Wert Zeit
    F2_end_zeit = F2_zeit_ueberlauf + temp;             // Endwert Zeit
    if((TIM10->SR & TIM_SR_UIF) && (temp < 0x8000)) {   // evtl. Ueberlauf T10 noch offen?
      F2_end_zeit += 0x10000;                           // dann korrigieren
    }  
    set_pin(GPIOC, TRIG2, 0);                           // Trigger-Ausgang wieder auf '0'
    F2_mess_status = AUSWERTEN;                         // Daten fertig fuer Auswertung
  }
  if(TIM9->SR & TIM_SR_UIF) {                           // bei berlufen von T9
     TIM9->SR = ~TIM_SR_UIF;                            // flag wieder loeschen
     F2_ueberlauf += 0x10000;                           // und berlufe zhlen
  }
  __ISB();                                    // gelschte Flags abwarten
}


/* direkte Messung mit internen Timern mit max. 8 Stellen/s
Timer9 zhlt die externen Impulse
Timer10 mit die Zeit mit internem Takt
*/
void F2_messung(uint8_t init_timer)
{
static uint8_t  init = 0;
static char s[MAX_ZEICHEN];
  if(init_timer) {
    init = 0;                                   // Anzeige + Messung initialisieren
    init_T6();                                  // T6 fr Messdauer in 1 ms Schritten
    init_t9();                                  // T9 als Ereignisszhler mit capture-modus starten
    init_t10();                                 // T10 als Referenz-Zhler
    RCC->AHB1ENR |= RCC_AHB1ENR_GPIOCEN;        // PortC aktivieren
    GPIOC->MODER    |= OUT << TRIG2*2;          // als Trigger-Ausgang
    GPIOC->OSPEEDR  |= V_HIGH_SPEED << TRIG2*2; // mit steilen Flanken
  }
    if(F2_mess_dauer >= F2_timeout) {           // bei fehlenden Eingangsimpulsen
      set_pin(GPIOC, TRIG2, 0);                 // Trigger-Ausgang wieder auf '0'
      sende_string("kein Signal");
      init = 0;                                 // neu initialisieren
    }  
    if(!init) {
      F2_mess_status  = MESSEN;                 // mit MESSEN Beginnen
      F2_mess_dauer =  F2_messzeit;             // und 1. Messung sofort abschlieen
      F2_messwert_vorhanden  = 0;               // dabei 1. Ergebnis verwerfen
      init = 1;
    }  

    if(F2_mess_status == MESSEN  && F2_mess_dauer >= F2_messzeit) {
      F2_mess_status = AUSLESEN;                // Zhler bei nchstem Eingangsimpuls auslesen
      set_pin(GPIOC, TRIG2, 1);                 // Trigger-Ausgang auf '1'
    }
  
    if(F2_mess_status == AUSWERTEN) {           // Ereignisse und Zeitpunkt ermittelt
      F2_mess_zeit = F2_end_zeit - F2_start_zeit; // Zeit-Differenz bilden
      F2_start_zeit = F2_end_zeit;
      
      F2_mess_ereignisse = F2_end_ereignis - F2_start_ereignis;  // Impuls-Differenz
      F2_start_ereignis = F2_end_ereignis;      // fuers naechste Intervall

      F2_mess_status = MESSEN;                  // neu starten
      F2_mess_dauer = 0;                        // und Messzeit abwarten
      if(F2_messwert_vorhanden) {               // Ergebnis berechnen und anzeigen/ausgeben
        F2_ref_frequenz = SystemCoreClock + F2_ref_offset;  // immer aktuellen Wert verwenden
        eff_zeit2 = (double)(F2_mess_zeit ) / F2_ref_frequenz;  // Zeit in Sekunden
        F2_ergebnis = ((double)F2_mess_ereignisse) / eff_zeit2; // Frequenz berechnen
        sprintf(s,"%13.7E",F2_ergebnis);        // 8-stelliges Ergebnis im Exp-Format
        sende_string(s);
      } else 
        F2_messwert_vorhanden = 1;              // sperre wieder aufheben
    }
}  

