/* 
Reziproke Frequenzmessung mit STM32F730 mit 10-stelliger Auflsung.
Eingangsfrequenbereich: 0,1 Hz - 100 MHz

Die Funktion der reziproken Frequenzmessung wird als bekannt vorausgesetzt.
Eine nhere Beschreibung findet sich im Programm "F2_messung.c".

Hier wird nun nicht die vorgegebene Messzeit abgewartet, um dann eine
Auswertung durchzufhren, sondern es wird nach Mglichkeit jedes einzelne
Intervall erfasst. Dazu wird die Anforderung zur Erfassung der Intervalle nicht
mehr einzeln, sondern per Timer7 ausgelst. Dieser setzt mit "F_SAMPLES" (500 kHz)
den Ausgang "TRIG2". Bei kleinerer Eingangsfrequenz wird jedes Intervall bewertet,
bei hheren Frequenzen soviele Perioden, wie in 2 s (synchronisiert) eintreffen.
Da Erfassung und Auswertung per ISR stattfinden, mu fr die restlichen 
Funktionen noch Rechenzeit brig bleiben, was hier mit dem Signal am "KONTROLL_AUSGANG"
(PB3) kontrolliert werden kann. 
Derzeit werden ca. 55% fr die ISR von Timer9 verbraucht.

Bei der Erfassung eines Intervalls werden die Summen der Einzelwerte von Ereignissen
und Zeit (reg_xi, reg_yi), die Summe aus diesen Einzelsummen (reg_sum_xi, reg_sum_yi)
und die Summen der Produkte aus reg_xi*reg_xi und reg_xi*reg_yi (reg_sum_xi2, reg_sum_xiyi)
gebildet. Ferner wird die Anzahl der Ereignisse (reg_n) gezhlt.
Nach Ablauf der Messzeit, kann aus diesen Werten mit linearer Regression die
verfeinerte Steigung der Geraden errechnet werden.

Die Grundauflsung ist 8-stellig, steigt jedoch mit Wurzel aus reg_n mit zunehmender
Eingangsfrequenz. Bei 1 s Messzeit sind folgende Auflsungen zu erwarten:

0,1 Hz - 50 Hz:    8-stellig
50 Hz - 5 kHz:     9-stellig
5 kHz - 500 kHz:  10-stellig
>= 500 kHz:       11-stellig

Erhht man die Messzeit auf 10 s erhlt man eine weitere Stelle Auflsung hinzu.
Zu beachten ist, da fr derartige Ergebnisse uerst stabile Signale fr
Eingangs- und Referenzfrequenz bentigt werden.

Die Messwerte werden ber USART3 ausgegeben: "sende_sting(s)"

2020-10-16
Michael Nowak
http://www.mino-elektronik.de
Alle Angaben ohne Gewaehr !
*/

#define F2_MESSUNG

#include <stdio.h>
#include <string.h>
#include "stm32f730xx.h"
extern void sende_string(char *s);    // Ausgabe ber USART3

#define DEF_F2_TAKT       216000000   // 216 MHz
#define DEF_F2_MESSZEIT   1000        // 1 s
#define DEF_TIMEOUT       12000       // nach 12 s
#define MAX_ZEICHEN       50          // fuer Ergebnis String

#define PRIO_TIMER        4           // hohe Prioritt fr T9 und T10
#define PRIO_TIM6         14          // unkritisch
#define PRIO_TIM7         10          // < PRIO_TIMER sonst unkritisch
#define AFR_TIM9          3           // fr T9  lt. Datenblatt
#define AFR_TIM10         3           // fr T10 lt. Datenblatt

#define F_SAMPLES         500000      // max. Messrate festlegen
#define KONTROLL_AUSGANG  3           // PB3 fr Laufzeitmessung

#define BIT(x) (1<<x)
enum gpio_modus {IN, OUT, AF, ANALOG};
enum gpio_speed {LOW_SPEED, MEDIUM_SPEED, HIGH_SPEED, V_HIGH_SPEED};
enum gpio_pullup {NO_PULL, PULL_UP, PULL_DOWN};

#define CLR_MODER         3         // Lschmaske
#define SET_GPIO          0x1       // untere Bits
#define RESET_GPIO        0x10000   // obere Bits

// PortB
#define TIM10_CAPTURE2    8
// PortC
#define TRIG2             0
// PortE
#define TIM9_ZAEHLER2     5
#define TIM9_CAPTURE2     6

enum ablauf {MESSEN=0, FERTIG};     // die Phasen der Messung
 
double              F2_ergebnis,                    // in Hz
                    eff_zeit2;                      // eff. Messzeit in s
double              F2_ref_frequenz = DEF_F2_TAKT;  // Vorgabewerte
int32_t             F2_ref_offset = 0,              // fr Abgleich von F2_ref
                    F2_messzeit = DEF_F2_MESSZEIT,
                    F2_timeout = DEF_TIMEOUT;

volatile uint8_t    F2_messwert_vorhanden,          // gueltige Messungen
                    F2_mess_status;                 // Ablaufsteuerung
                                   
volatile uint32_t   F2_mess_dauer;                  // in ms
uint32_t            F2_start_ereignis,              // Impulse zu Beginn der Messung
                    F2_start_zeit,                  // rel. Zeit: Beginn der Messung
                    F2_zeit_ueberlauf,
                    F2_end_zeit,                    // Zeitpunkt der Auswertung
                    F2_mess_zeit,                   // genaue Zeit der Messung
                    F2_mess_ereignisse,             // Impulse der Messung
                    F2_ueberlauf; 
volatile uint32_t   F2_end_ereignis;                // Impulse am Ende der Messung

// Variablen fr Regressionsberechnung
uint32_t  reg_n;                          // Anzahl Ereignisse bei Regression
volatile  uint32_t reg_ereignisse;        // und Endwert zur weiteren Auswertung
uint64_t  reg_xi, reg_yi;                 // summe fortlaufender Einzelwerte       
uint64_t  reg_sum_xi, reg_sum_yi;         // Summe aller Intervalle

double    reg_sum_xiyi,                   // Summe xi * yi
          reg_sum_xi2;                    // Summe xi
double    reg_xq, reg_yq;                 // Mittelwerte xi/n und yi/n

void loesche_regression(void)
{
  reg_n = 0;                              // und alle Werte der Regressionsberechnung
  reg_xi = reg_yi = 0;                    // lschen
  reg_sum_xi = reg_sum_yi = 0;
  reg_sum_xiyi = reg_sum_xi2 = 0;
  reg_xq = reg_yq = 0;
}

// direktes Setzen oder Ruecksetzen eines Port-Pins
void set_pin(GPIO_TypeDef* GPIOx, uint8_t number, char set) {
  if(set) {
    GPIOx->BSRR = (SET_GPIO << number);       // IO-Bit setzen
  } else {
    GPIOx->BSRR = (RESET_GPIO << number);     // IO-Bit loeschen
  }
}


// 1 kHz ISR mit T6 erzeugen
void init_t6(void)
{
static uint8_t init;
  if(!init) {
    init = 1;
    RCC->DCKCFGR1 |= RCC_DCKCFGR1_TIMPRE;     // Timer auf max. Takt
    RCC->APB1ENR |= RCC_APB1ENR_TIM6EN;
    TIM6->ARR = 1000-1;                       // Teiler fr 1 kHz
    TIM6->PSC = SystemCoreClock/1000000-1;    // Vorteiler auf 1 MHz
    TIM6->CR1 = 0x1;                          // starten
    TIM6->DIER = 0x1;                         // und Interrupt freigeben
    
    NVIC_EnableIRQ(TIM6_DAC_IRQn);
    NVIC_SetPriority(TIM6_DAC_IRQn,PRIO_TIM6);// geringe Prioritaet
  }    
} 

// Messdauer im 1 ms Raster inkrementieren
void TIM6_DAC_IRQHandler(void)                // Aufruf mit 1 kHz
{
  TIM6->SR = 0;                               // flag loeschen
  F2_mess_dauer++;                            // Mesdauer in ms-Raster
  __ISB();                                    // gelschte Flags abwarten
} 

// 500 kHz ISR mit T7 erzeugen
void init_t7(void)
{
static uint8_t init;
  if(!init) {
    init = 1;
    RCC->DCKCFGR1 |= RCC_DCKCFGR1_TIMPRE;     // Timer auf max. Takt
    RCC->AHB1ENR |= (RCC_AHB1ENR_GPIOBEN);    // PortB aktivieren
    GPIOB->MODER &= ~(CLR_MODER << (KONTROLL_AUSGANG * 2));// MODER PB3 = 0
    GPIOB->MODER |= (OUT << (KONTROLL_AUSGANG * 2));       // PB3 auf Ausgang         
    RCC->APB1ENR |= RCC_APB1ENR_TIM7EN;
    TIM7->ARR = SystemCoreClock/F_SAMPLES-1;  // max. Abfragefrequenz
    TIM7->CR1 = 0x1;                          // starten
    TIM7->DIER = 0x1;                         // und Interrupt freigeben
    
    NVIC_EnableIRQ(TIM7_IRQn);
    NVIC_SetPriority(TIM7_IRQn,PRIO_TIM7);  // geringe Prioritaet
  }    
} 

void TIM7_IRQHandler(void)                  // Aufruf mit F_SAMPLES
{
    TIM7->SR = 0;
    set_pin(GPIOC, TRIG2, 1);               // neue Werte anfordern
  __ISB();                                  // gelschte Flags abwarten
}

void init_t9(void)
{
  RCC->APB2ENR |= RCC_APB2ENR_TIM9EN;
	RCC->AHB1ENR |= (RCC_AHB1ENR_GPIOEEN);    // PortE aktivieren
  TIM9->CR1 = 0;                            // T9 stoppen, falls schon aktiv
  
// alternative Funktion der Zaehlereingaenge einschalten  
  GPIOE->MODER |= (AF << (TIM9_ZAEHLER2 * 2) |
                   AF << (TIM9_CAPTURE2 * 2));                       
  
// Pullup zu den Eingaengen einschalten
  GPIOE->PUPDR |= (PULL_UP << (TIM9_ZAEHLER2 * 2) |
                   PULL_UP << (TIM9_CAPTURE2 * 2));                       

// PE5 als Zhlereingang, PE6 als capture-input zu T9
  GPIOE->AFR[0] |=  (AFR_TIM9 << (TIM9_ZAEHLER2 * 4) | 
                     AFR_TIM9 << (TIM9_CAPTURE2 * 4));                   


  NVIC_EnableIRQ(TIM1_BRK_TIM9_IRQn);                   // Int-T9 aktivieren
  NVIC_SetPriority(TIM1_BRK_TIM9_IRQn,PRIO_TIMER);

  TIM9->CCMR1 = TIM_CCMR1_CC2S_0;                       // capture zuordnung  IC2 -> TI2
  TIM9->CCER =  TIM_CCER_CC2E;                          // pos. Flanken an IC2 freigeben
  TIM9->SMCR |= TIM_SMCR_TS_2 + TIM_SMCR_TS_0;          // ext. Signal von TI1
  TIM9->SMCR |= TIM_SMCR_SMS_2 + TIM_SMCR_SMS_1 + TIM_SMCR_SMS_0;   // Zhlereingang von TI1

  TIM9->DIER |= TIM_DIER_CC2IE + TIM_DIER_UIE;          // mit interrupts
  TIM9->CR1 |=  TIM_CR1_CEN;                            // T9 starten  
}

void init_t10(void)
{
  RCC->APB2ENR |= RCC_APB2ENR_TIM10EN;
	RCC->AHB1ENR |= (RCC_AHB1ENR_GPIOBEN);    // PortB aktivieren
  TIM10->CR1 = 0;                           // T10 stoppen, falls schon aktiv

  GPIOB->MODER |= (AF << (TIM10_CAPTURE2 * 2));         // alternative Funktion         
  GPIOB->PUPDR |= (PULL_UP << (TIM10_CAPTURE2 * 2));    // Pullup einschalten

// PB8 als capture-input zu T10
  GPIOB->AFR[1] |=  (AFR_TIM10 << ((TIM10_CAPTURE2-8) * 4));                   


  NVIC_EnableIRQ(TIM1_UP_TIM10_IRQn);                   // Int-T10 aktivieren
  NVIC_SetPriority(TIM1_UP_TIM10_IRQn,PRIO_TIMER);
  
  TIM10->CCMR1 =  TIM_CCMR1_CC1S_0;                     // capture zuordnung  IC1 -> TI1
  TIM10->CCER = TIM_CCER_CC1E;                          // pos. Flanken an IC1 freigeben

  TIM10->DIER |= TIM_DIER_UIE;                          // mit interrupts
  TIM10->CR1 |= TIM_CR1_CEN;                            // T10 starten
}

// ISR zu Timer10
// berlufe bei der Zeitmessung zhlen
void TIM1_UP_TIM10_IRQHandler(void)
{
  if(TIM10->SR & TIM_SR_UIF) {                          // bei berlufen von T10
     TIM10->SR = ~TIM_SR_UIF;                           // flag wieder loeschen
     F2_zeit_ueberlauf += 0x10000;                      // und berlufe zhlen
  }
  __ISB();                                    // gelschte Flags abwarten
}

// ISR zu Timer9
// synchron zu einem Eingangsimpuls Ereignisse und Zeitpunkt ermitteln
// anschlieend Auswertung mit lin. Regression
// ferner berlaufe bei der Ereigniszhlung zhlen
void TIM1_BRK_TIM9_IRQHandler(void)
{
uint32_t temp;
double reg_divisor;
  if(TIM9->SR & TIM_SR_CC2IF) {                         // bei Capture-Ereignis
    set_pin(GPIOB, KONTROLL_AUSGANG, 1);    // zur Kontrolle
    TIM9->SR = ~TIM_SR_CC2IF;                           // flag wieder loeschen
    temp = TIM9->CCR2;                                  // Capture-Wert Ereignisse
    F2_end_ereignis = F2_ueberlauf + temp;              // Endwert Ereignisse
    if((TIM9->SR & TIM_SR_UIF) && (temp < 0x8000))      // evtl. Ueberlauf T9 noch offen?
      F2_end_ereignis += 0x10000;                       // nur, wenn capture-int + overflow-int gleichzeitig !

    temp = TIM10->CCR1;                                 // Capture-Wert Zeit
    F2_end_zeit = F2_zeit_ueberlauf + temp;             // Endwert Zeit
    if((TIM10->SR & TIM_SR_UIF) && (temp < 0x8000)) {   // evtl. Ueberlauf T10 noch offen?
      F2_end_zeit += 0x10000;                           // dann korrigieren
    }  

// jedes Intervall berechnen und weiterverarbeiten
    F2_mess_zeit = F2_end_zeit - F2_start_zeit; // Zeit-Differenz bilden
    F2_start_zeit = F2_end_zeit;
      
    F2_mess_ereignisse = F2_end_ereignis - F2_start_ereignis;  // Impuls-Differenz
    F2_start_ereignis = F2_end_ereignis;        // fuers naechste Intervall

// Anzahl und Einzelwerte fr die Regression verarbeiten      
    reg_n++;                          // anzahl der Einzelwerte
    reg_xi += F2_mess_ereignisse;     // Einzelwerte Addieren
    reg_yi += F2_mess_zeit;           // dto
    reg_sum_xi += reg_xi;             // Summe aus Summe der Einzelwerte
    reg_sum_yi += reg_yi;             // dto
    reg_sum_xiyi += reg_xi*reg_yi;    // Summe der Produkte
    reg_sum_xi2  += reg_xi*reg_xi;    // Summes der Quadrate
    
// Auswertung nach Ablauf der Messdauer starten    
    if(F2_mess_status == MESSEN &&            // wenn Auswertung aktiv
       F2_mess_dauer >= F2_messzeit) {        // und Mindestzeit erreicht
      F2_ref_frequenz = SystemCoreClock + F2_ref_offset;  // immer aktuellen Wert verwenden
      reg_xq = (double)reg_sum_xi/(double)reg_n;
      reg_yq = (double)reg_sum_yi/(double)reg_n;
      reg_divisor = (reg_sum_xi2 - reg_n * (reg_xq * reg_xq));
      if(reg_n >= 2 && reg_divisor != 0.0) {  // Regression erst ab zwei Werten          
        F2_ergebnis = F2_ref_frequenz / ((reg_sum_xiyi - reg_n * (reg_xq * reg_yq)) / reg_divisor);
      } else {                                // wenn nur ein Ereignis in reg_xi
        F2_ergebnis = ((double)reg_xi/(double)reg_yi)*F2_ref_frequenz;  // Berechnung ohne Regression
      }  
// Endwerte fr Debugger      
      reg_ereignisse = reg_n;
        
      F2_mess_dauer = 0;                      // neue Messzeit starten
      loesche_regression();
      F2_mess_status = FERTIG;                // Daten fertig fuer Anzeige
    }
    set_pin(GPIOC, TRIG2,0);                  // alte Anforderung lschen, TIM7 setzt neue Anforderung 
    set_pin(GPIOB, KONTROLL_AUSGANG, 0);      // zur Kontrolle
  }

// ganz zum Schluss noch die berlufe bercksichtigen
  if(TIM9->SR & TIM_SR_UIF) {                 // bei berlufen von T9
     TIM9->SR = ~TIM_SR_UIF;                  // flag wieder loeschen
     F2_ueberlauf += 0x10000;                 // und berlufe zhlen
  }
  __ISB();                                    // gelschte Flags abwarten
}


/* direkte Messung mit internen Timern mit max. 10 Stellen/s
Timer9 zhlt die externen Impulse
Timer10 mit die Zeit mit internem Takt
*/
void F2_messung(uint8_t init_timer)
{
static uint8_t  init = 0;
static char s[MAX_ZEICHEN];
  if(init_timer) { 
    init = 0;                                   // Anzeige + Messung initialisieren
    init_t6();                                  // T6 fr Messdauer in 1 ms Schritten
    init_t7();                                  // TRIG2 aktivieren, max. Messrate vorgeben
    init_t9();                                  // T9 als Ereignisszhler mit capture-modus starten
    init_t10();                                 // T10 als Referenz-Zhler
    RCC->AHB1ENR |= RCC_AHB1ENR_GPIOCEN;        // PortC aktivieren
    GPIOC->MODER    |= OUT << TRIG2*2;          // als Trigger-Ausgang
    GPIOC->OSPEEDR  |= V_HIGH_SPEED << TRIG2*2; // mit steilen Flanken
  }
//  while(1) {    
    if(!init) {
      F2_mess_status  = MESSEN;                 // mit MESSEN beginnen
      F2_mess_dauer =  F2_messzeit;             // und 1. Messung sofort abschlieen
      F2_messwert_vorhanden  = 0;               // dabei 1. Ergebnis verwerfen
      init = 1;
    }
    
    if(F2_mess_dauer >= F2_timeout) {
      set_pin(GPIOC, TRIG2, 0);                 // Trigger-Ausgang wieder auf '0'
      sende_string("kein Signal");
      init = 0;                                 // neu initialisieren
    }  
 
    if(F2_mess_status == FERTIG) {              // Regressionsberechnung fertig
      if(F2_messwert_vorhanden) {               // Ergebnis anzeigen/ausgeben
        sprintf(s,"%16.10E",F2_ergebnis);       // 11-stelliges Ergebnis im Exp-Format
        sende_string(s);
      } else {
        F2_messwert_vorhanden = 1;              // sperre wieder aufheben
      }  
      F2_mess_status  = MESSEN;                 // neue Auswertung zulassen
    }
//  }
}  

